"""
Module implementing some clustering algorithms that work in parallel on batches of data.
"""

import heat as ht
import torch
from heat.cluster._kcluster import _KCluster
from heat.core.dndarray import DNDarray
from warnings import warn
from math import log


from typing import Union, Tuple, TypeVar

self = TypeVar("self")

"""
Auxiliary single-process functions and base class for batch-parallel k-clustering
"""


def _initialize_plus_plus(X, n_clusters, p, random_state=None):
    """
    Auxiliary function: single-process k-means++/k-medians++ initialization in pytorch
    p is the norm used for computing distances
    """
    if random_state is not None:
        torch.manual_seed(random_state)
    idxs = torch.zeros(n_clusters, dtype=torch.long, device=X.device)
    idxs[0] = torch.randint(0, X.shape[0], (1,))
    for i in range(1, n_clusters):
        dist = torch.cdist(X, X[idxs[:i]], p=p)
        dist = torch.min(dist, dim=1)[0]
        idxs[i] = torch.multinomial(dist, 1)
    return X[idxs]


def _kmex(X, p, n_clusters, init, max_iter, tol, random_state=None):
    """
    Auxiliary function: single-process k-means and k-medians in pytorch
    p is the norm used for computing distances: p=2 implies k-means, p=1 implies k-medians.
    p should be 1 (k-medians) or 2 (k-means). For other choice of p, we proceed as for p=2 and hope for the best.
    (note: kmex stands for kmeans and kmedians)
    """
    if random_state is not None:
        torch.manual_seed(random_state)
    if isinstance(init, torch.Tensor):
        if init.shape != (n_clusters, X.shape[1]):
            raise ValueError("if a torch tensor, init must have shape (n_clusters, n_features).")
        centers = init
    elif init == "++":
        centers = _initialize_plus_plus(X, n_clusters, p, random_state)
    elif init == "random":
        idxs = torch.randint(0, X.shape[0], (n_clusters,))
        centers = X[idxs]
    else:
        raise ValueError(
            "init must be a torch tensor with initial centers, string '++', or 'random'."
        )
    for _ in range(max_iter):
        dist = torch.cdist(X, centers, p=p)
        labels = torch.argmin(dist, dim=1)
        # update centers
        centers_old = centers.clone()
        for i in range(n_clusters):
            if (labels == i).any():
                if p == 1:
                    centers[i] = torch.median(X[labels == i], dim=0)[0]
                else:
                    centers[i] = torch.mean(X[labels == i], dim=0)
            else:
                # if a cluster is empty, we leave its center unchanged
                pass
        # check if tolerance is reached
        if torch.allclose(centers, centers_old, atol=tol):
            break
    return centers, _ + 1


def _parallel_batched_kmex_predict(X, centers, p):
    """
    Auxiliary function: predict labels for parallel_batched_kmex
    """
    dist = torch.cdist(X, centers, p=p)
    return torch.argmin(dist, dim=1).reshape(-1, 1)


class _BatchParallelKCluster(ht.ClusteringMixin, ht.BaseEstimator):
    """
    Base class for batch parallel k-clustering
    """

    def __init__(
        self,
        p: int,
        n_clusters: int,
        init: str,
        max_iter: int,
        tol: float,
        random_state: Union[int, None],
        n_procs_to_merge: Union[int, None],
    ):  # noqa: D107
        if not isinstance(n_clusters, int):
            raise TypeError(f"n_clusters must be int, but was {type(n_clusters)}")
        if n_clusters <= 0:
            raise ValueError(f"n_clusters must be positive, but was {n_clusters}")
        if not isinstance(max_iter, int):
            raise TypeError(f"max_iter must be int, but was {type(max_iter)}")
        if max_iter <= 0:
            raise ValueError(f"max_iter must be positive, but was {max_iter}")
        if not isinstance(tol, float):
            raise TypeError(f"tol must be float, but was {type(tol)}")
        if tol <= 0:
            raise ValueError(f"tol must be positive, but was {tol}")

        if not isinstance(random_state, int) and random_state is not None:
            raise TypeError(f"random_state must be int or None, but was {type(random_state)}")
        if not isinstance(n_procs_to_merge, int) and n_procs_to_merge is not None:
            raise TypeError(f"procs_to_merge must be int or None, but was {type(n_procs_to_merge)}")
        if n_procs_to_merge is not None and n_procs_to_merge <= 1:
            raise ValueError(
                f"If an integer, procs_to_merge must be > 1, but was {n_procs_to_merge}."
            )

        self.n_clusters = n_clusters
        self._init = init
        self.max_iter = max_iter
        self.tol = tol
        self.random_state = random_state
        self.n_procs_to_merge = n_procs_to_merge

        # in-place properties
        if not (p == 1 or p == 2):
            warn(
                "p should be 1 (k-Medians) or 2 (k-Means). For other choice of p, we proceed as for p=2 and hope for the best.",
                UserWarning,
            )
        self._p = p
        self._cluster_centers = None
        self._n_iter = None
        self._functional_value = None

    @property
    def cluster_centers_(self) -> DNDarray:
        """
        Returns the coordinates of the cluster centers. Returns None if fit has not been called yet.
        """
        return self._cluster_centers

    @property
    def n_iter_(self) -> Tuple[int]:
        """
        Returns the number of iterations run. Returns None if fit has not been called yet.
        The output is of the form (n_iter_local, n_iter_global), where n_iter_local is the number of iterations of the local k-means/k-medians,
        and n_iter_global is the number of iterations of the global k-means/k-medians;
        """
        return self._n_iter

    @property
    def functional_value_(self) -> float:
        """
        Returns the value of the K-Clustering functional. Returns None if fit and predict have not been called yet.
        """
        return self._functional_value

    def fit(self, x: DNDarray):
        """
        Computes the centroid of the clustering algorithm to fit the data ``x``.

        Parameters
        ----------
        x : DNDarray
            Training instances to cluster. Shape = (n_samples, n_features). It must hold x.split=0.

        """
        if not isinstance(x, DNDarray):
            raise TypeError(f"input needs to be a ht.DNDarray, but was {type(x)}")
        if not x.ndim == 2:
            raise ValueError(f"input needs to be 2D, but was {x.ndim}D")
        if x.split != 0:
            raise ValueError(
                f"input needs to be split along the sample axis, but was split along {x.split}"
            )

        # local k-clustering
        local_random_state = None if self.random_state is None else self.random_state + x.comm.rank
        centers_local, n_iters_local = _kmex(
            x.larray,
            self._p,
            self.n_clusters,
            self._init,
            self.max_iter,
            self.tol,
            local_random_state,
        )

        # hierarchical approach to obtail "global" cluster centers from the "local" centers
        procs_to_merge = self.n_procs_to_merge if self.n_procs_to_merge is not None else x.comm.size
        current_procs = [i for i in range(x.comm.size)]
        current_comm = x.comm
        local_comm = current_comm.Split(current_comm.rank // procs_to_merge, x.comm.rank)

        level = 1
        while len(current_procs) > 1:
            if x.comm.rank in current_procs and local_comm.size > 1:
                # create array to collect centers from all processes of the process group of at most n_procs_to_merge processes
                if local_comm.rank == 0:
                    gathered_centers_local = torch.zeros(
                        (local_comm.size, self.n_clusters, x.shape[1]),
                        device=x.larray.device,
                        dtype=x.larray.dtype,
                    )
                else:
                    gathered_centers_local = torch.empty(
                        0, device=x.larray.device, dtype=x.larray.dtype
                    )
                # gather centers from all processes of the process group of at most n_procs_to_merge processes
                local_comm.Gather(centers_local, gathered_centers_local, root=0, axis=0)
                # k-clustering on the gathered centers
                if local_comm.rank == 0:
                    gathered_centers_local = gathered_centers_local.reshape(-1, x.shape[1])
                    centers_local, n_iters_local_new = _kmex(
                        gathered_centers_local,
                        self._p,
                        self.n_clusters,
                        self._init,
                        self.max_iter,
                        self.tol,
                        local_random_state,
                    )
                    del gathered_centers_local
                    n_iters_local += n_iters_local_new

            # update: determine processes to be active at next "merging" level, create new communicator and split it into groups for gathering
            current_procs = [
                current_procs[i] for i in range(len(current_procs)) if i % procs_to_merge == 0
            ]
            if len(current_procs) > 1:
                new_group = x.comm.group.Incl(current_procs)
                current_comm = x.comm.Create_group(new_group)
                if x.comm.rank in current_procs:
                    local_comm = ht.communication.MPICommunication(
                        current_comm.Split(current_comm.rank // procs_to_merge, x.comm.rank)
                    )
            level += 1

        # broadcast the final centers to all processes
        x.comm.Bcast(centers_local, root=0)

        self._cluster_centers = DNDarray(
            centers_local,
            (self.n_clusters, x.shape[1]),
            dtype=x.dtype,
            device=x.device,
            comm=x.comm,
            split=None,
            balanced=True,
        )
        self._n_iter = n_iters_local
        return self

    def predict(self, x: DNDarray):
        """
        Predict the closest cluster each sample in ``x`` belongs to.

        In the vector quantization literature, :func:`cluster_centers_` is called the code book and each value returned by
        predict is the index of the closest code in the code book.

        Parameters
        ----------
        x : DNDarray
            New data to predict. Shape = (n_samples, n_features)
        """
        # input sanitation
        if not isinstance(x, DNDarray):
            raise TypeError(f"input needs to be a ht.DNDarray, but was {type(x)}")
        if not x.ndim == 2:
            raise ValueError(f"input needs to be 2D, but was {x.ndim}D")
        if x.split != 0:
            raise ValueError(
                f"input needs to be split along the sample axis, but was split along {x.split}"
            )
        if self._cluster_centers is None:
            raise RuntimeError("fit needs to be called before predict")
        if x.shape[1] != self._cluster_centers.shape[1]:
            raise ValueError(
                f"input needs to have {self._cluster_centers.shape[1]} features, but has {x.shape[1]}"
            )

        local_labels = _parallel_batched_kmex_predict(
            x.larray, self._cluster_centers.larray, self._p
        )
        labels = DNDarray(
            local_labels,
            gshape=(x.shape[0], 1),
            dtype=ht.int32,
            device=x.device,
            comm=x.comm,
            split=x.split,
            balanced=x.balanced,
        )
        if self._p == 2:
            self._functional_value = (
                torch.norm(
                    x.larray - self._cluster_centers.larray[local_labels, :].squeeze(), p="fro"
                )
                ** 2
            )
        else:
            self._functional_value = torch.norm(
                x.larray - self._cluster_centers.larray[local_labels, :].squeeze(), p=self._p, dim=1
            ).sum()
        x.comm.Allreduce(ht.communication.MPI.IN_PLACE, self._functional_value)
        self._functional_value = self._functional_value.item()
        return labels


"""
Actual classes for batch parallel K-means and K-medians
"""


class BatchParallelKMeans(_BatchParallelKCluster):
    r"""
    Batch-parallel K-Means clustering algorithm from Ref. [1].
    The input must be a ``DNDarray`` of shape `(n_samples, n_features)`, with split=0 (i.e. split along the sample axis).
    This method performs K-Means clustering on each batch (i.e. on each process-local chunk) of data individually and in parallel.
    After that, all centroids from the local K-Means are gathered and another instance of K-means is performed on them in order to determine the final centroids.
    To improve scalability of this approach also on a large number of processes, this procedure can be applied in a hierarchical manner using the parameter `n_procs_to_merge`.

    Attributes
    ----------
    n_clusters : int
        The number of clusters to form as well as the number of centroids to generate.
    init : str
        Method for initialization for local and global k-means:
        - ‘k-means++’ : selects initial cluster centers for the clustering in a smart way to speed up convergence [2].
        - ‘random’: choose k observations (rows) at random from data for the initial centroids. (Not implemented yet)
    max_iter : int
        Maximum number of iterations of the local/global k-means algorithms.
    tol : float
        Relative tolerance with regards to inertia to declare convergence, both for local and global k-means.
    random_state : int
        Determines random number generation for centroid initialization.
    n_procs_to_merge : int
        Number of processes to merge after each iteration of the local k-means. If None, all processes are merged after each iteration.


    References
    ----------
    [1] Rasim M. Alguliyev, Ramiz M. Aliguliyev, Lyudmila V. Sukhostat, Parallel batch k-means for Big data clustering, Computers & Industrial Engineering, Volume 152 (2021). https://doi.org/10.1016/j.cie.2020.107023.
    """

    def __init__(
        self,
        n_clusters: int = 8,
        init: str = "k-means++",
        max_iter: int = 300,
        tol: float = 1e-4,
        random_state: int = None,
        n_procs_to_merge: int = None,
    ):  # noqa: D107
        if not isinstance(init, str):
            raise TypeError(f"init must be str, but was {type(init)}")
        else:
            if init == "k-means++":
                _init = "++"
            elif init == "random":
                raise NotImplementedError(
                    "random initialization for batch parallel k-means is currently not supported due to instable behaviour of the algorithm. Use init='k-means++' instead."
                )
            else:
                raise ValueError(f"init must be 'k-means++' or 'random', but was {init}")
        super().__init__(
            p=2,
            n_clusters=n_clusters,
            init=_init,
            max_iter=max_iter,
            tol=tol,
            random_state=random_state,
            n_procs_to_merge=n_procs_to_merge,
        )
        self.init = init


class BatchParallelKMedians(_BatchParallelKCluster):
    r"""
    Batch-parallel K-Medians clustering algorithm, in analogy to the K-means algorithm from Ref. [1].
    This requires data to be given as DNDarray of shape (n_samples, n_features) with split=0 (i.e. split along the sample axis).
    The idea of the method is to perform the classical K-Medians on each batch of data (i.e. on each process-local chunk of data) individually and in parallel.
    After that, all centroids from the local K-Medians are gathered and another instance of K-Medians is performed on them in order to determine the final centroids.
    To improve scalability of this approach also on a range number of processes, this procedure can be applied in a hierarchical manor using the parameter n_procs_to_merge.

    Attributes
    ----------
    n_clusters : int
        The number of clusters to form as well as the number of centroids to generate.
    init : str
        Method for initialization for local and global k-medians:
        - ‘k-medians++’ : selects initial cluster centers for the clustering in a smart way to speed up convergence [2].
        - ‘random’: choose k observations (rows) at random from data for the initial centroids. (Not implemented yet)
    max_iter : int
        Maximum number of iterations of the local/global k-Medians algorithms.
    tol : float
        Relative tolerance with regards to inertia to declare convergence, both for local and global k-Medians.
    random_state : int
        Determines random number generation for centroid initialization.
    n_procs_to_merge : int
        Number of processes to merge after each iteration of the local k-Medians. If None, all processes are merged after each iteration.


    References
    ----------
    [1] Rasim M. Alguliyev, Ramiz M. Aliguliyev, Lyudmila V. Sukhostat, Parallel batch k-means for Big data clustering, Computers & Industrial Engineering, Volume 152 (2021). https://doi.org/10.1016/j.cie.2020.107023.
    """

    def __init__(
        self,
        n_clusters: int = 8,
        init: str = "k-medians++",
        max_iter: int = 300,
        tol: float = 1e-4,
        random_state: int = None,
        n_procs_to_merge: int = None,
    ):  # noqa: D107
        if not isinstance(init, str):
            raise TypeError(f"init must be str, but was {type(init)}")
        else:
            if init == "k-medians++":
                _init = "++"
            elif init == "random":
                raise NotImplementedError(
                    "random initialization for batch parallel k-medians is currently not supported due to instable behaviour of the algorithm. Use init='k-medians++' instead."
                )
            else:
                raise ValueError(f"init must be 'k-medians++' or 'random', but was {init}")
        super().__init__(
            p=1,
            n_clusters=n_clusters,
            init=_init,
            max_iter=max_iter,
            tol=tol,
            random_state=random_state,
            n_procs_to_merge=n_procs_to_merge,
        )
        self.init = init



"""
QR decomposition of (distributed) 2-D ``DNDarray``s.
"""

import collections
import torch
from typing import Tuple

from ..dndarray import DNDarray
from .. import factories
from .. import communication
from ..types import float32, float64

__all__ = ["qr"]


def qr(
    A: DNDarray,
    mode: str = "reduced",
    procs_to_merge: int = 2,
) -> Tuple[DNDarray, DNDarray]:
    r"""
    Calculates the QR decomposition of a 2D ``DNDarray``.
    Factor the matrix ``A`` as *QR*, where ``Q`` is orthonormal and ``R`` is upper-triangular.
    If ``mode = "reduced``, function returns ``QR(Q=Q, R=R)``, if ``mode = "r"`` function returns ``QR(Q=None, R=R)``

    Parameters
    ----------
    A : DNDarray of shape (M, N)
        Array which will be decomposed. So far only 2D arrays with datatype float32 or float64 are supported
        For split=0, the matrix must be tall skinny, i.e. the local chunks of data must have at least as many rows as columns.
    mode : str, optional
        default "reduced" returns Q and R with dimensions (M, min(M,N)) and (min(M,N), N), respectively.
        "r" returns only R, with dimensions (min(M,N), N).
    procs_to_merge : int, optional
        This parameter is only relevant for split=0 and determines the number of processes to be merged at one step during the so-called TS-QR algorithm.
        The default is 2. Higher choices might be faster, but will probably result in higher memory consumption. 0 corresponds to merging all processes at once.
        We only recommend to modify this parameter if you are familiar with the TS-QR algorithm (see the references below).

    Notes
    -----
    The distribution schemes of ``Q`` and ``R`` depend on that of the input ``A``.

        - If ``A`` is distributed along the columns (A.split = 1), so will be ``Q`` and ``R``.

        - If ``A`` is distributed along the rows (A.split = 0), ``Q`` too will have  `split=0`, but ``R`` won't be distributed, i.e. `R. split = None` and a full copy of ``R`` will be stored on each process.

    Note that the argument `calc_q` allowed in earlier Heat versions is no longer supported; `calc_q = False` is equivalent to `mode = "r"`.
    Unlike ``numpy.linalg.qr()``, `ht.linalg.qr` only supports ``mode="reduced"`` or ``mode="r"`` for the moment, since "complete" may result in heavy memory usage.

    Heats QR function is built on top of PyTorchs QR function, ``torch.linalg.qr()``, using LAPACK (CPU) and MAGMA (CUDA) on
    the backend. For split=0, tall-skinny QR (TS-QR) is implemented, while for split=1 a block-wise version of stabilized Gram-Schmidt orthogonalization is used.

    References
    -----------
    Basic information about QR factorization/decomposition can be found at, e.g.:

        - https://en.wikipedia.org/wiki/QR_factorization,

        - Gene H. Golub and Charles F. Van Loan. 1996. Matrix Computations (3rd Ed.).

    For an extensive overview on TS-QR and its variants we refer to, e.g.,

        - Demmel, James, et al. “Communication-Optimal Parallel and Sequential QR and LU Factorizations.” SIAM Journal on Scientific Computing, vol. 34, no. 1, 2 Feb. 2012, pp. A206–A239., doi:10.1137/080731992.
    """
    if not isinstance(A, DNDarray):
        raise TypeError(f"'A' must be a DNDarray, but is {type(A)}")
    if not isinstance(mode, str):
        raise TypeError(f"'mode' must be a str, but is {type(mode)}")
    if mode not in ["reduced", "r"]:
        if mode == "complete":
            raise NotImplementedError(
                "QR decomposition with 'mode'='complete' is not supported by heat yet. \n Please open an issue on GitHub if you require this feature. \n For now, you can use 'mode'='reduced' or 'r' instead."
            )
        elif mode == "raw":
            raise NotImplementedError(
                "QR decomposition with 'mode'='raw' is neither supported by Heat nor by PyTorch. \n"
            )
        else:
            raise ValueError(f"'mode' must be 'reduced' (default) or 'r', but is {mode}")
    if not isinstance(procs_to_merge, int):
        raise TypeError(f"procs_to_merge must be an int, but is currently {type(procs_to_merge)}")
    if procs_to_merge < 0 or procs_to_merge == 1:
        raise ValueError(
            f"procs_to_merge must be 0 (for merging all processes at once) or at least 2, but is currently {procs_to_merge}"
        )
    if procs_to_merge == 0:
        procs_to_merge = A.comm.size

    if A.ndim != 2:
        raise ValueError(
            f"Array 'A' must be 2 dimensional, buts has {A.ndim} dimensions. \n Please open an issue on GitHub if you require QR for batches of matrices similar to PyTorch."
        )
    if A.dtype not in [float32, float64]:
        raise TypeError(f"Array 'A' must have a datatype of float32 or float64, but has {A.dtype}")

    QR = collections.namedtuple("QR", "Q, R")

    if not A.is_distributed():
        # handle the case of a single process or split=None: just PyTorch QR
        Q, R = torch.linalg.qr(A.larray, mode=mode)
        R = DNDarray(
            R,
            gshape=R.shape,
            dtype=A.dtype,
            split=A.split,
            device=A.device,
            comm=A.comm,
            balanced=True,
        )
        if mode == "reduced":
            Q = DNDarray(
                Q,
                gshape=Q.shape,
                dtype=A.dtype,
                split=A.split,
                device=A.device,
                comm=A.comm,
                balanced=True,
            )
        else:
            Q = None
        return QR(Q, R)

    if A.split == 1:
        # handle the case that A is split along the columns
        # here, we apply a block-wise version of (stabilized) Gram-Schmidt orthogonalization
        # instead of orthogonalizing each column of A individually, we orthogonalize blocks of columns (i.e. the local arrays) at once

        lshapes = A.lshape_map[:, 1]
        lshapes_cum = torch.cumsum(lshapes, 0)
        nprocs = A.comm.size

        if A.shape[0] >= A.shape[1]:
            last_row_reached = nprocs
            k = A.shape[1]
        else:
            last_row_reached = min(torch.argwhere(lshapes_cum >= A.shape[0]))[0]
            k = A.shape[0]

        if mode == "reduced":
            Q = factories.zeros(A.shape, dtype=A.dtype, split=1, device=A.device, comm=A.comm)

        R = factories.zeros((k, A.shape[1]), dtype=A.dtype, split=1, device=A.device, comm=A.comm)
        R_shapes = torch.hstack(
            [
                torch.zeros(1, dtype=torch.int32, device=A.device.torch_device),
                torch.cumsum(R.lshape_map[:, 1], 0),
            ]
        )

        A_columns = A.larray.clone()

        for i in range(last_row_reached + 1):
            # this loop goes through all the column-blocks (i.e. local arrays) of the matrix
            # this corresponds to the loop over all columns in classical Gram-Schmidt
            if i < nprocs - 1:
                k_loc_i = min(A.shape[0], A.lshape_map[i, 1])
                Q_buf = torch.zeros(
                    (A.shape[0], k_loc_i), dtype=A.larray.dtype, device=A.device.torch_device
                )

            if A.comm.rank == i:
                # orthogonalize the current block of columns by utilizing PyTorch QR
                Q_curr, R_loc = torch.linalg.qr(A_columns, mode="reduced")
                if i < nprocs - 1:
                    Q_buf = Q_curr
                if mode == "reduced":
                    Q.larray = Q_curr
                r_size = R.larray[R_shapes[i] : R_shapes[i + 1], :].shape[0]
                R.larray[R_shapes[i] : R_shapes[i + 1], :] = R_loc[:r_size, :]

            if i < nprocs - 1:
                # broadcast the orthogonalized block of columns to all other processes
                req = A.comm.Ibcast(Q_buf, root=i)
                req.Wait()

            if A.comm.rank > i:
                # subtract the contribution of the current block of columns from the remaining columns
                R_loc = Q_buf.T @ A_columns
                A_columns -= Q_buf @ R_loc
                r_size = R.larray[R_shapes[i] : R_shapes[i + 1], :].shape[0]
                R.larray[R_shapes[i] : R_shapes[i + 1], :] = R_loc[:r_size, :]

        if mode == "reduced":
            Q = Q[:, :k].balance()
        else:
            Q = None

        return QR(Q, R)

    if A.split == 0:
        # implementation of TS-QR for split = 0
        # check that data distribution is reasonable for TS-QR (i.e. tall-skinny matrix with also tall-skinny local chunks of data)
        if A.lshape_map[:, 0].max().item() < A.shape[1]:
            raise ValueError(
                "A is split along the rows and the local chunks of data are rectangular with more rows than columns. \n Applying TS-QR in this situation is not reasonable w.r.t. runtime and memory consumption. \n We recomment to split A along the columns instead. \n In case this is not an option for you, please open an issue on GitHub."
            )

        current_procs = [i for i in range(A.comm.size)]
        current_comm = A.comm
        local_comm = current_comm.Split(current_comm.rank // procs_to_merge, A.comm.rank)
        Q_loc, R_loc = torch.linalg.qr(A.larray, mode=mode)
        R_loc = R_loc.contiguous()  # required for all the communication ops lateron
        if mode == "reduced":
            leave_comm = current_comm.Split(current_comm.rank, A.comm.rank)

        level = 1
        while len(current_procs) > 1:
            if A.comm.rank in current_procs and local_comm.size > 1:
                # create array to collect the R_loc's from all processes of the process group of at most n_procs_to_merge processes
                shapes_R_loc = local_comm.gather(R_loc.shape[0], root=0)
                if local_comm.rank == 0:
                    gathered_R_loc = torch.zeros(
                        (sum(shapes_R_loc), R_loc.shape[1]),
                        device=R_loc.device,
                        dtype=R_loc.dtype,
                    )
                    counts = list(shapes_R_loc)
                    displs = torch.cumsum(
                        torch.tensor([0] + shapes_R_loc, dtype=torch.int32), 0
                    ).tolist()[:-1]
                else:
                    gathered_R_loc = torch.empty(0, device=R_loc.device, dtype=R_loc.dtype)
                    counts = None
                    displs = None
                # gather the R_loc's from all processes of the process group of at most n_procs_to_merge processes
                local_comm.Gatherv(R_loc, (gathered_R_loc, counts, displs), root=0, axis=0)
                # perform QR decomposition on the concatenated, gathered R_loc's to obtain new R_loc
                if local_comm.rank == 0:
                    previous_shape = R_loc.shape
                    Q_buf, R_loc = torch.linalg.qr(gathered_R_loc, mode=mode)
                    R_loc = R_loc.contiguous()
                else:
                    Q_buf = torch.empty(0, device=R_loc.device, dtype=R_loc.dtype)
                if mode == "reduced":
                    if local_comm.rank == 0:
                        Q_buf = Q_buf.contiguous()
                    scattered_Q_buf = torch.empty(
                        R_loc.shape if local_comm.rank != 0 else previous_shape,
                        device=R_loc.device,
                        dtype=R_loc.dtype,
                    )
                    # scatter the Q_buf to all processes of the process group
                    local_comm.Scatterv((Q_buf, counts, displs), scattered_Q_buf, root=0, axis=0)
                del gathered_R_loc, Q_buf

            # for each process in the current processes, broadcast the scattered_Q_buf of this process
            # to all leaves (i.e. all original processes that merge to the current process)
            if mode == "reduced" and leave_comm.size > 1:
                try:
                    scattered_Q_buf_shape = scattered_Q_buf.shape
                except UnboundLocalError:
                    scattered_Q_buf_shape = None
                scattered_Q_buf_shape = leave_comm.bcast(scattered_Q_buf_shape, root=0)
                if scattered_Q_buf_shape is not None:
                    # this is needed to ensure that only those Q_loc get updates that are actually part of the current process group
                    if leave_comm.rank != 0:
                        scattered_Q_buf = torch.empty(
                            scattered_Q_buf_shape, device=Q_loc.device, dtype=Q_loc.dtype
                        )
                    leave_comm.Bcast(scattered_Q_buf, root=0)
                # update the local Q_loc by multiplying it with the scattered_Q_buf
            try:
                Q_loc = Q_loc @ scattered_Q_buf
                del scattered_Q_buf
            except UnboundLocalError:
                pass

            # update: determine processes to be active at next "merging" level, create new communicator and split it into groups for gathering
            current_procs = [
                current_procs[i] for i in range(len(current_procs)) if i % procs_to_merge == 0
            ]
            if len(current_procs) > 1:
                new_group = A.comm.group.Incl(current_procs)
                current_comm = A.comm.Create_group(new_group)
                if A.comm.rank in current_procs:
                    local_comm = communication.MPICommunication(
                        current_comm.Split(current_comm.rank // procs_to_merge, A.comm.rank)
                    )
                if mode == "reduced":
                    leave_comm = A.comm.Split(A.comm.rank // procs_to_merge**level, A.comm.rank)
            level += 1
        # broadcast the final R_loc to all processes
        R_gshape = (A.shape[1], A.shape[1])
        if A.comm.rank != 0:
            R_loc = torch.empty(R_gshape, dtype=R_loc.dtype, device=R_loc.device)
        A.comm.Bcast(R_loc, root=0)
        R = DNDarray(
            R_loc,
            gshape=R_gshape,
            dtype=A.dtype,
            split=None,
            device=A.device,
            comm=A.comm,
            balanced=True,
        )
        if mode == "r":
            Q = None
        else:
            Q = DNDarray(
                Q_loc,
                gshape=A.shape,
                dtype=A.dtype,
                split=0,
                device=A.device,
                comm=A.comm,
                balanced=True,
            )
        return QR(Q, R)
